#!/usr/bin/perl
# decode code
# V0.4.5
#
# Copyright (c) 2012-2013, Sasaji
#use utf8;
use strict;
use Getopt::Std;
use File::Basename;

our %opts = ();
our $charset_type = "0";	# 
our $through_mode = 1;		# 
our $cr_code = "\n";

our %char_code_table = ();
our %char_code_index = ();

our $app_path=dirname($0)."/";

getopts("i:n:h",\%opts);

if (exists($opts{"h"})) {
	&usage();
	exit 0;
}
if ($opts{"i"} ne "") {
	if ($opts{"i"} eq "0") {
		$through_mode = 0;
		$charset_type = "Interace";
	} elsif ($opts{"i"} eq "1") {
		$through_mode = 0;
		$charset_type = "NonInterace";
	} elsif ($opts{"i"} eq "2") {
		$through_mode = 0;
		$charset_type = "Interace(Wide)";
	} else {
		&usage();
		exit 0;
	}
}
if ($opts{"n"} ne "") {
	if (lc($opts{"n"}) eq "cr") {
		$cr_code = "\r";
	} elsif (lc($opts{"n"}) eq "lf") {
		$cr_code = "\n";
	} elsif (lc($opts{"n"}) eq "crlf") {
		$cr_code = "\r\n";
	} else {
		&usage();
		exit 0;
	}
}
if (@ARGV <= 0) {
	&usage();
	exit 0;
}

if (set_char_code_table(\%char_code_table)) {
	exit 1;
}

foreach my $infile (@ARGV) {
	&main($infile);
}

# $k=getc(STDIN);

exit 0;

#####
sub usage() {
	my $prog_name=basename($0);

	print STDERR "$prog_name [options] file ...\n";
	print STDERR " Decode char code."."\n";
	print STDERR '  options:'."\n";
	print STDERR '  -i 0   : convert file(s) using interace mode charset(utf-8).'."\n";
	print STDERR '           zenkaku kata/hira -> hankaku kata/hira'."\n";
	print STDERR '  -i 1   : convert file(s) using non-interace mode charset(utf-8).'."\n";
	print STDERR '  -i 2   : convert file(s) using interace mode charset(utf-8).'."\n";
	print STDERR '           wide chars are converted to narrow char'."\n";
	print STDERR '  -n cr  : output CR as newline .'."\n";
	print STDERR '  -n lf  : output LF as newline .'."\n";
	print STDERR '  -n crlf: output CR+LF as newline .'."\n";
	print STDERR '           default newline code is LF.'."\n";
}

#####

sub main() {
	my($infile) = @_;
	my $fh;

	my $buf="";
	my $buf1="";
	my $buflen = 0;

	if (! open($fh,$infile)) {
		print STDERR $infile.' cannot open.'."\n";
		return;
	}
	binmode($fh);
	binmode(STDOUT);

	my %parse_state;
	my $val = 0;

	&init_data(\%parse_state);

	if ($through_mode == 1) {
		print $cr_code;
	}
	my $len = 0;
	do {
		$buflen = read($fh,$buf1,2048);
		$buf .= $buf1;
		$len = length($buf);
		while($len >= 768 || $buflen <= 0) {
			$parse_state{file_line}++;
			$buf =~ s/^[\s\r\n]*//;
			last if ($buf eq "");

			if (&decode_data(\$buf, \%parse_state) != 0) {
				$buf =~ s/^.*?[\r\n]+//;
			}
			$len = length($buf);
		}
	} while ($buf ne "" || $buflen > 0);
	close($fh);

#	print "end\n";
}

#################
sub init_data() {
	my($rs)=@_;

	$rs->{mode} = 0;
	$rs->{file_line} = 0;		# 

	$rs->{subpos} = 0;			# 

	$rs->{body} = "";			# 

}

# parse per line
sub decode_data() {
	my($rline, $rs)=@_;
	my $statement="";
	my $new_state="";

	my $len_state = 0;

	$rs->{subpos} = 0;			# 

	$rs->{body} = "";			# 

	# parse statement
	$len_state = length($$rline);
	$new_state = "";
	while($rs->{subpos} < $len_state) {
		my $ch = substr($$rline, $rs->{subpos}, 1);
		my $find = 0;

		if ($ch =~ /^[\r\n]/) {
			last;
		}
#		print STDERR " ".unpack("H2",$ch);

		if ($through_mode == 0) {
			# convert char code
			my $val = unpack("C", $ch);
			if (exists($char_code_table{$charset_type}{$val})) {
				# match
				$new_state .= $char_code_table{$charset_type}{$val}{val};
				$find = 1;
				$rs->{subpos}++;
			}
		}
		if ($find != 1) {
			$new_state .= $ch;
			$rs->{subpos}++;
		}
	}

	$rs->{body} = $new_state.$cr_code;

	print $rs->{body};

	$$rline = substr($$rline, $rs->{subpos});

	return 0;
}


sub set_char_code_table()
{
	my($tbl)=@_;
	my $fh;
	my $key;
	my @arr;
	my $charset = "";

	if (! open($fh,$app_path."char_code.dat")) {
		print STDERR "char_code.dat open error.\n";
		return 1;
	}
	while (my $line = readline($fh)) {
		$line =~ s/^\s*(.+)\s*$/$1/;

		next if ($line =~ /^#/);

		chomp($line);

		# charset type code
		if ($line =~ /^\[(.+)\]$/) {
			$charset = $1;
			if (!exists($tbl->{$charset})) {
				$tbl->{$charset} = {};
			}
#		printf STDERR "%s\n",$charset;
		} else {
			next if ($charset eq "");

			@arr = split(/\s*,\s*/,$line);

			next if ($arr[0] eq "");

			next if ($arr[2] ne "" && $arr[2] ne ">");

			$key = hex($arr[0]);

			next if ($key > 255);

			$arr[1] =~ s/^\'(.+)\'$/$1/;

			$tbl->{$charset}{$key}{val} =$arr[1];

#		printf STDERR "%d %s\n",$key,$tbl->{$charset}{$key}{val};
		}

	}
	close($fh);

	return 0;
}
