#!/usr/bin/perl
# convert L3 BASIC from ascii code to internal code
# V0.4.5
#
# Copyright (c) 2012-2013, Sasaji
#use utf8;
use strict;
use Getopt::Std;
use File::Basename;

our %opts = ();
our $charset_type = "0";	# 
our $through_mode = 1;		# 
our $basic_type = "";

our %basic_code_table = ();
our %basic_code_index = ();
our %char_code_table = ();
our %char_code_index = ();

our $app_path=dirname($0)."/";

getopts("d:i:n:h",\%opts);

if (defined($opts{"h"})) {
	&usage();
	exit 0;
}
if ($opts{"d"} ne "") {
	if ($opts{"d"} eq "3") {
		$basic_type = "DISK BASIC(3inch)";
	} elsif ($opts{"d"} eq "5") {
		$basic_type = "DISK BASIC(5inch)";
	} else {
		&usage();
		exit 0;
	}
}
if ($opts{"i"} ne "") {
	if ($opts{"i"} eq "0") {
		$through_mode = 0;
		$charset_type = "Interace";
	} elsif ($opts{"i"} eq "1") {
		$through_mode = 0;
		$charset_type = "NonInterace";
	} elsif ($opts{"i"} eq "2") {
		$through_mode = 0;
		$charset_type = "Interace(Wide)";
	} else {
		&usage();
		exit 0;
	}
}
if (@ARGV <= 0) {
	&usage();
	exit 0;
}

if (set_basic_code_table(\%basic_code_table)) {
	exit 1;
}
set_basic_code_index(\%basic_code_table, \%basic_code_index);

if (set_char_code_table(\%char_code_table)) {
	exit 1;
}
foreach my $key (keys %char_code_table) {
	$char_code_index{$key}={} if (! defined($char_code_index{$key}));
	set_char_code_index($char_code_table{$key}, $char_code_index{$key});
}

foreach my $infile (@ARGV) {
	&main($infile);
}

# $k=getc(STDIN);

exit 0;

#####
sub usage() {
	my $prog_name=basename($0);

	print STDERR "$prog_name [options] file ...\n";
	print STDERR " Encode to internal code for L3 BASIC from text file(s)."."\n";
	print STDERR '  options:'."\n";
	print STDERR '  -d 3   : parse as "DISK BASIC(3inch)".'."\n";
	print STDERR '  -d 5   : parse as "DISK BASIC(5inch)".'."\n";
	print STDERR '  -i 0   : convert file(s) using interace mode charset(utf-8).'."\n";
	print STDERR '           zenkaku kata/hira -> hankaku kata/hira'."\n";
	print STDERR '  -i 1   : convert file(s) using non-interace mode charset(utf-8).'."\n";
	print STDERR '  -i 2   : convert file(s) using interace mode charset(utf-8).'."\n";
	print STDERR '           wide chars are converted to narrow char'."\n";
}

#####

sub main() {
	my($infile) = @_;
	my $fh;

	my $buf="";
	my $buf1="";
	my $buflen = 0;

	if (! open($fh,$infile)) {
		print STDERR $infile.' cannot open.'."\n";
		return;
	}
	binmode($fh);

	my %parse_state;
	my $val = 0;

	&init_data(\%parse_state);

	binmode(STDOUT);
	print "\xff\xff\xff";	# header
	my $len = 0;
	do {
		$buflen = read($fh,$buf1,2048);
		$buf .= $buf1;
		$len = length($buf);
		while($len >= 768 || $buflen <= 0) {
			$parse_state{file_line}++;
			$buf =~ s/^[\s\r\n]*//;
			last if ($buf eq "");

			if (&encode_data(\$buf, \%parse_state) != 0) {
				$buf =~ s/^.*?[\r\n]+//;
			}
			$len = length($buf);
		}
	} while ($buf ne "" || $buflen > 0);
	print "\x00\x00";	# footer
	close($fh);

#	print "end\n";
}

#################
sub init_data() {
	my($rs)=@_;

	$rs->{mode} = 0;
	$rs->{file_line} = 0;		# 

	$rs->{addr_next} = 0x4796;	# 
#	$rs->{addr_next} = 0x1372;	# 
	$rs->{line_num} = 0;		# 
	$rs->{subpos} = 0;			# 

	$rs->{quoted} = 0;			# 
	$rs->{data_area} = 0;		# 
	$rs->{comment_area} = 0;	# 
	$rs->{variable_area} = 0;	# 

	$rs->{body} = "";			# 

}

# parse per line
sub encode_data() {
	my($rline, $rs)=@_;
	my $statement="";
	my $cnv_state="";
	my $new_state="";
	my $cnv_len = 0;

	my $len_state = 0;

	$rs->{subpos} = 0;			# 

	$rs->{quoted} = 0;			# 
	$rs->{data_area} = 0;		# 
	$rs->{comment_area} = 0;	# 
	$rs->{variable_area} = 0;	# 

	$rs->{body} = "";			# 

	# convert char
	$len_state = length($$rline);
	$cnv_state = "";
	my $pos = 1;
	while($rs->{subpos} < $len_state) {
		my $ch = substr($$rline, $rs->{subpos}, 1);
		my $find = 0;

		if ($ch =~ /^[\r\n]/) {
			last;
		}

		if ($through_mode == 0) {
			# convert char code
			if (exists($char_code_index{$charset_type}{$ch})) {
				foreach my $cmd (@{$char_code_index{$charset_type}{$ch}}) {
					my $cmdlen = length($cmd->{key});
					my $cmdpos = index($$rline, $cmd->{key}, $rs->{subpos});
					if ($cmdpos == $rs->{subpos}) {
						# match
						$cnv_state .= $cmd->{data}{val};
						$find = 1;
						$rs->{subpos} += $cmdlen;
						$pos += length($cmd->{data}{val});
						last;
					} 
				}
			}
			if ($find != 1) {
				my $val = unpack("C", $ch);
				if ($val >= 0x80) {
					print STDERR 'line:'.$rs->{file_line}.' pos:'.$pos.' cannot convert char.'."\n";
				}
			}
		}
		if ($find != 1) {
			$cnv_state .= $ch;
			$rs->{subpos}++;
			$pos++;
		}
	}

	# line_num statement...\n
	if ($cnv_state =~ s/^(\d+)\s//) {
		$rs->{line_num} = $1;
	} else {
		print STDERR "parse error.(line:".$rs->{file_line}." [".substr($$rline,0,20)."])\n";
		return 1;
	}

	# parse statement
	$cnv_len = 0;
	$len_state = length($cnv_state);
	$new_state = "";
	while($cnv_len < $len_state) {
		my $ch = substr($cnv_state, $cnv_len, 1);
		my $find = 0;
		if ($ch =~ /^[\r\n]/) {
			last;
		}

		if ($ch eq "\x22") {
			$rs->{quoted} = 1 - $rs->{quoted};
		}

		if ($rs->{variable_area} > 0) {
			if (($ch !~ /^[0-9a-zA-Z]/) || ($rs->{variable_area} > 16)) {
				# end of values area
				$rs->{variable_area} = 0;
			} else {
				$new_state .= uc($ch);
				$cnv_len++;
				$rs->{variable_area}++;
				$find = 1;
			}
		}

		# find command
		if ($rs->{comment_area} == 0
		 && $rs->{data_area} == 0
		 && $rs->{quoted} == 0
		 && $rs->{variable_area} == 0
		 && exists($basic_code_index{$ch})) {
			foreach my $cmd (@{$basic_code_index{$ch}}) {
				# find a BASIC sentence
				my $cmdlen = length($cmd->{key});
				my $cmdpos = index(uc($cnv_state), uc($cmd->{key}), $cnv_len);
				if ($cmdpos == $cnv_len) {
					# match
					if ($cmd->{data}{attr} =~ /colon/) {
						$new_state .= ":";
					}

					if ($cmd->{data}{val} >= 256) {
						$new_state .= pack("C",0xff);
					}
					$new_state .= pack("C",($cmd->{data}{val} & 0xff));

					if ($cmd->{data}{attr} =~ /comment/) {
						$rs->{comment_area} = 1;
					}
					if ($cmd->{data}{attr} =~ /data/) {
						$rs->{data_area} = 1;
					}

					$find = 1;
					$cnv_len += $cmdlen;
					last;
				} 
			}
			if ($find != 1) {
				# non BASIC sentence
				if ($ch =~ /^[a-zA-Z]/) {
					# is variable
					$rs->{variable_area} = 1;

					$new_state .= uc($ch);
					$cnv_len++;
					$find = 1;
				} else {
					my $val = unpack("C", $ch);
					if ($val >= 0x80) {
						print STDERR 'line:'.$rs->{file_line}.' pos:'.($cnv_len+1).' invalid char exists.'."\n";
					}
				}
			}
		}
		if ($find != 1) {
			$new_state .= $ch;
			$cnv_len++;
		}
	}
	$rs->{addr_next} += length($new_state) + 5;

	my $addr_next = pack("n",$rs->{addr_next}) . "";
	my $line_num  = pack("n",$rs->{line_num}) . "";

	$rs->{body} = $addr_next.$line_num.$new_state."\x00";

	print $rs->{body};

	$$rline = substr($$rline, $rs->{subpos});

	return 0;
}


sub set_basic_code_index()
{
	my($tbl, $idx)=@_;

	foreach my $key (sort keys %{$tbl}) {
		my $ch = substr($key, 0, 1);
		$idx->{$ch} = [] if (!exists($idx->{$ch}));
		unshift(@{$idx->{$ch}}, { key => $key, data => $tbl->{$key} });

#		printf STDERR "%s\n",$key;
	}
	return $idx;
}


sub set_basic_code_table()
{
	my($tbl)=@_;
	my $fh;
	my $key;
	my @arr;
	my $type = "";
	my $target = 0;

	if (! open($fh,$app_path."basic_code.dat")) {
		print STDERR "basic_code.dat open error.\n";
		return 1;
	}
	while (my $line = readline($fh)) {
		$line =~ s/^\s*(.+)\s*$/$1/;

		next if ($line =~ /^#/);

		chomp($line);

		# basic type
		if ($line =~ /^\[(.*)\]$/) {
			$type = $1;
			if ($type eq "" || $type eq $basic_type) {
				$target = 1;
			} else {
				$target = 0;
			}
		} else {
			next if ($target != 1);

			@arr = split(/\s*,\s*/,$line);

			next if ($arr[0] eq "");

			$key = $arr[1];

			$tbl->{$key}{val} =hex($arr[0]);
			$tbl->{$key}{attr}=$arr[2];
		}
#		printf STDERR "%s %d\n",$key,$tbl->{$key}{val};
	}
	close($fh);

	return 0;
}

sub set_char_code_index()
{
	my($tbl, $idx)=@_;

	foreach my $key (sort keys %{$tbl}) {
		my $ch = substr($key, 0, 1);
		$idx->{$ch} = [] if (!exists($idx->{$ch}));
		unshift(@{$idx->{$ch}}, { key => $key, data => $tbl->{$key} });

#		printf STDERR "%s\n",$key;
	}
	return $idx;
}

sub set_char_code_table()
{
	my($tbl)=@_;
	my $fh;
	my $key;
	my $val;
	my @arr;
	my $charset = "";

	if (! open($fh,$app_path."char_code.dat")) {
		print STDERR "char_code.dat open error.\n";
		return 1;
	}
	while (my $line = readline($fh)) {
		$line =~ s/^\s*(.+)\s*$/$1/;

		next if ($line =~ /^#/);

		chomp($line);

		# charset type code
		if ($line =~ /^\[(.+)\]$/) {
			$charset = $1;
			if (!exists($tbl->{$charset})) {
				$tbl->{$charset} = {};
			}
#		printf STDERR "%s\n",$charset;
		} else {
			next if ($charset eq "");

			@arr = split(/\s*,\s*/,$line);

			next if ($arr[0] eq "");

			next if ($arr[2] ne "" && $arr[2] ne "<");

			$arr[1] =~ s/^\'(.+)\'$/$1/;
			$key = $arr[1];
			$val = hex($arr[0]);
			if ($val >= 256) {
				$tbl->{$charset}{$key}{val} = pack("C2",($val >> 8) & 0xff,$val & 0xff);
			} else {
				$tbl->{$charset}{$key}{val} = pack("C",$val & 0xff);
			}
#		printf STDERR "%s %s\n",$key,$tbl->{$charset}{$key}{val};
		}

	}
	close($fh);

	return 0;
}
